from __future__ import absolute_import

import os
import sys

import flask
from flask._compat import reraise
import pathlib
from werkzeug.utils import import_string

from keg.extensions import lazy_gettext as _


# sentinal object
class NotGiven(object):
    pass


def ensure_dirs(newdir, mode=NotGiven):
    """
        A "safe" verision of Path.makedir(..., parents=True) that will only create the directory
        if it doesn't already exist.  We also manually create parents so that mode is set
        correctly.  Python docs say that mode is ignored when using Path.mkdir(..., parents=True)
    """
    if mode is NotGiven:
        mode = flask.current_app.config['KEG_DIR_MODE']
    dpath = pathlib.Path(newdir)
    if dpath.is_dir():
        return
    if dpath.is_file():
        raise OSError(_("A file with the same name as the desired"
                        " directory, '{dpath}', already exists.", dpath=dpath))

    ensure_dirs(dpath.parent, mode=mode)
    dpath.mkdir(mode)


def reraise_lastexc():
    exc_type, exc_value, tb = sys.exc_info()
    reraise(exc_type, exc_value, tb)


class ClassProperty(property):
    """A decorator that behaves like @property except that operates
    on classes rather than instances.
    """
    def __init__(self, fget, *arg, **kw):
        self.ignore_set = kw.pop('ignore_set', False)
        self.__doc__ = fget.__doc__
        super(ClassProperty, self).__init__(fget, *arg, **kw)

    def __get__(self, obj, cls):  # noqa
        return self.fget(cls)

    def __set__(self, cls, value):  # noqa
        if self.fset is None and not self.ignore_set:
            raise AttributeError(_("can't set attribute"))
        if not self.ignore_set:
            self.fset(cls, value)


classproperty = ClassProperty


class HybridMethod(object):
    """
        A decorator which allows definition of a Python object method with both
        instance-level and class-level behavior:

            Class Bar:

                @hybridmethod
                def foo(self, rule, **options):
                    # this is used in an instance context

                @foo.classmethod
                def foo(cls, rule, **options):
                    # this is used in class context

    """

    def __init__(self, func, cm_func=None):
        self.instance_func = func
        self.classmethod(cm_func or func)

    def __get__(self, instance, owner):
        if instance is None:
            return self.cm_func.__get__(owner, owner.__class__)
        else:
            return self.instance_func.__get__(instance, owner)

    def classmethod(self, cm_func):
        """Provide a modifying decorator that is used as a classmethod decorator."""

        self.cm_func = cm_func
        if not self.cm_func.__doc__:
            self.cm_func.__doc__ = self.instance_func.__doc__
        return self


hybridmethod = HybridMethod


def pymodule_fpaths_to_objects(fpaths):
    """
        Takes an iterable of file paths reprenting possible python modules and will return an
        iterable of tuples with the file path along with the contents of that file if the file
        exists.
    """
    retval = []
    for fpath in fpaths:
        path = pathlib.Path(fpath)
        if path.exists():
            pymodule_globals = {}
            with open(fpath) as fo:
                exec(compile(fo.read(), fpath, 'exec'), pymodule_globals)
            retval.append((fpath, pymodule_globals))
    return retval


def app_environ_get(app_import_name, key, default=None):
    # App names often have periods and it is not possibe to export an
    # environment variable with a period in it.
    app_name = app_import_name.replace('.', '_').upper()
    environ_key = '{}_{}'.format(app_name, key.upper())

    return os.environ.get(environ_key, default)


def visit_modules(dotted_paths, base_path=None):
    for path in dotted_paths:
        if path.startswith('.') and base_path is not None:
            path = base_path + path
        import_string(path)



from __future__ import absolute_import

import pytest

from keg.app import Keg, KegAppError


class TestInit(object):

    @pytest.mark.skip("Fails when flask version is >1.0")
    def test_init_called_twice_error(self):
        with pytest.raises(KegAppError, match=r'init\(\) already called on this instance'):
            app = Keg(__name__)
            app.init(use_test_profile=True)
            app.init(use_test_profile=True)


class TestRouteDecorator(object):
    class TRDApp(Keg):
        import_name = __name__ + ':TRDApp'

    class TRDApp2(Keg):
        import_name = __name__ + ':TRDApp2'

    def test_class_method(self):
        @self.TRDApp.route('/foo')
        def foo():
            pass

        @self.TRDApp.route('/bar')
        def bar():
            pass

        @self.TRDApp2.route('/baz')
        def baz():
            pass

        assert len(self.TRDApp._routes) == 2
        assert len(self.TRDApp2._routes) == 1



from __future__ import absolute_import

import importlib
import warnings

import flask
from flask.config import ConfigAttribute
from six.moves import range
from werkzeug.datastructures import ImmutableDict

from keg.blueprints import keg as kegbp
import keg.cli
import keg.config
from keg.ctx import KegRequestContext
import keg.logging
import keg.signals as signals
from keg.extensions import lazy_gettext as _
from keg.templating import _keg_default_template_ctx_processor, AssetsExtension
from keg.utils import classproperty, visit_modules, hybridmethod
import keg.web


class KegAppError(Exception):
    pass


class Keg(flask.Flask):
    import_name = None
    use_blueprints = ()
    oauth_providers = ()
    keyring_enabled = ConfigAttribute('KEG_KEYRING_ENABLE')
    config_class = keg.config.Config
    logging_class = keg.logging.Logging
    keyring_manager_class = None

    _cli = None
    cli_loader_class = keg.cli.CLILoader

    db_enabled = False
    db_visit_modules = ['.model.entities']
    db_manager = None

    jinja_options = ImmutableDict(
        extensions=['jinja2.ext.autoescape', 'jinja2.ext.with_', AssetsExtension]
    )

    template_filters = ImmutableDict()
    template_globals = ImmutableDict()

    visit_modules = False

    _init_ran = False

    def __init__(self, import_name=None, *args, **kwargs):

        # flask requires an import name, so we should too.
        if import_name is None and self.import_name is None:
            raise KegAppError(_('Please set the "import_name" attribute on your app class or'
                                ' pass it into the app instance.'))

        # passed in value takes precedence
        import_name = import_name or self.import_name

        self.keyring_manager = None
        self._init_config = kwargs.pop('config', {})

        flask.Flask.__init__(self, import_name, *args, **kwargs)

    def make_config(self, instance_relative=False):
        """
            Needed for Flask <= 0.10.x so we can set the configuration class
            being used.  Once 0.11 comes out, Flask supports setting the config_class on the app.
        """
        root_path = self.root_path
        if instance_relative:
            root_path = self.instance_path
        return self.config_class(root_path, self.default_config)

    def init(self, config_profile=None, use_test_profile=False, config=None):
        if self._init_ran:
            raise KegAppError(_('init() already called on this instance'))
        self._init_ran = True

        self.init_config(config_profile, use_test_profile, config)
        self.init_logging()
        self.init_keyring()
        self.init_oath()
        self.init_error_handling()
        self.init_extensions()
        self.init_routes()
        self.init_blueprints()
        self.init_jinja()
        self.init_visit_modules()

        self.on_init_complete()
        signals.app_ready.send(self)
        signals.init_complete.send(self)

        # return self for easy chaining, i.e. app = MyKegApp().init()
        return self

    def on_init_complete(self):
        """ For subclasses to override """
        pass

    def init_config(self, config_profile, use_test_profile, config):
        init_config = self._init_config.copy()
        init_config.update(config or {})

        self.config.init_app(config_profile, self.import_name, self.root_path, use_test_profile)

        self.config.update(init_config)

        signals.config_ready.send(self)
        signals.config_complete.send(self)
        self.on_config_complete()

    def on_config_complete(self):
        """ For subclasses to override """
        pass

    def init_keyring(self):
        # do keyring substitution
        if self.keyring_enabled:
            from keg.keyring import Manager, keyring
            if keyring is None:
                warnings.warn(_('Keyring substitution is enabled, but the keyring package is not'
                                ' installed.  Please install the keyring package (pip install'
                                ' keyring) or disable keyring support by setting'
                                ' `KEG_KEYRING_ENABLE = False` in your configuration profile.'))
                return

            self.keyring_manager = Manager(self)
            self.keyring_manager.substitute(self.config)

    def init_extensions(self):
        self.init_db()

        # Keg components are essentially flask extensions, so they need to be set up during the
        # init process. Being extensions, it makes sense to trigger this here. But, they will best
        # be set up at the end of the init process, because other app-specific extensions may be
        # set up in `on_init_complete` for error handling/notifications and the like.
        signals.init_complete.connect(self.init_registered_components, sender=self)

    def db_manager_cls(self):
        from keg.db import DatabaseManager
        return DatabaseManager

    def init_db(self):
        if self.db_enabled:
            cls = self.db_manager_cls()
            self.db_manager = cls(self)

    def init_registered_components(self, app):
        # KEG_REGISTERED_COMPONENTS is presumed to be a set/list/iterable of dotted paths usable
        # for import. At the top level of the imported path, there should be a `__component__`
        # that takes the dotted path that was used for import (as an absolute parent for relative
        # imports) and has an init_app. Ideally, based on KegComponent.
        for comp_path in self.config.get('KEG_REGISTERED_COMPONENTS', set()):
            comp_module = importlib.import_module(comp_path)
            comp_object = getattr(comp_module, '__component__')
            comp_object.init_app(self, parent_path=comp_path)

    def init_blueprints(self):
        # TODO: probably want to be selective about adding our blueprint
        self.register_blueprint(kegbp)
        for blueprint in self.use_blueprints:
            self.register_blueprint(blueprint)

    def init_logging(self):
        self.logging = self.logging_class(self.config)
        self.logging.init_app()

    def init_error_handling(self):
        # handle status codes
        generic_errors = range(500, 506)
        for err in generic_errors:
            self.errorhandler(err)(self.handle_server_error)

        # utility to abort responses
        self.errorhandler(keg.web.ImmediateResponse)(keg.web.handle_immediate_response)

    def init_oath(self):
        # if no providers are listed, then we don't need to do anything else
        if not self.oauth_providers:
            return

        from keg.oauth import oauthlib, bp, manager
        self.register_blueprint(bp)
        oauthlib.init_app(self)
        manager.register_providers(self.oauth_providers)

    def init_jinja(self):
        self.jinja_env.filters.update(self.template_filters)

        # template_context_processors is supposed to be functions that return dictionaries where
        # the key is the name of the template variable and the value is the value.
        # First, add Keg defaults
        self.template_context_processors[None].append(_keg_default_template_ctx_processor)
        self.template_context_processors[None].append(lambda: self.template_globals)

    def init_visit_modules(self):
        if self.visit_modules:
            visit_modules(self.visit_modules, self.import_name)

    def handle_server_error(self, error):
        # send_exception_email()
        return '500 SERVER ERROR<br/><br/>{}'.format(_('administrators notified'))

    def request_context(self, environ):
        return KegRequestContext(self, environ)

    def _cli_getter(cls):  # noqa: first argument is not self in this context due to @classproperty
        if cls._cli is None:
            cal = cls.cli_loader_class(cls)
            cls._cli = cal.create_group()
        return cls._cli
    cli = classproperty(_cli_getter, ignore_set=True)

    @classmethod
    def environ_key(cls, key):
        # App names often have periods and it is not possibe to export an
        # environment variable with a period in it.
        name = cls.import_name.replace('.', '_').upper()
        return '{}_{}'.format(name, key.upper())

    @classmethod
    def testing_prep(cls, **config):
        """
            Make sure an instance of this class exists in a state that is ready for testing to
            commence.

            Trigger `signal.testing_run_start` the first time this method is called for an app.
        """
        # For now, do the import here so we don't have a hard dependency on WebTest
        from keg.testing import ContextManager
        if cls is Keg:
            raise TypeError(_('Don\'t use testing_prep() on Keg.  Create a subclass first.'))
        cm = ContextManager.get_for(cls)

        # if the context manager's app isn't ready, that means this will be the first time the app
        # is instantiated.  That seems like a good indicator that tests are just beginning, so it's
        # safe to trigger the signal.  We don't want the signal to fire every time b/c
        # testing_prep() can be called more than once per test run.
        trigger_signal = not cm.is_ready()
        cm.ensure_current(config)

        if trigger_signal:
            signals.testing_run_start.send(cm.app)

        return cm.app

    def make_shell_context(self):
        return {}

    @property
    def logger(self):
        return self.logging.app_logger

    @hybridmethod
    def route(self, rule, **options):
        """ Same as Flask.route() and will be used when in an instance context. """
        return super(Keg, self).route(rule, **options)

    @route.classmethod
    def route(cls, rule, **options):  # noqa
        """
            Enable .route() to be used in a class context as well.  E.g.:

            KegApp.route('/something'):
            def view_something():
                pass
        """
        def decorator(f):
            if not hasattr(cls, '_routes'):
                cls._routes = []
            cls._routes.append((f, rule, options))
            return f
        return decorator

    def init_routes(self):
        if not hasattr(self, '_routes'):
            return
        for func, rule, options in self._routes:
            # We follow the same logic here as Flask.route() decorator.
            endpoint = options.pop('endpoint', None)
            self.add_url_rule(rule, endpoint, func, **options)


@signals.init_complete.connect
def flask_wtf_bugfix(app):
    """ Needed until https://github.com/lepture/flask-wtf/issues/301 is fixed & released. """
    if not app.config.get('FIX_FLASK_WTF_BUG', True):
        return

    try:
        import flask_wtf
    except ImportError:
        return

    version = tuple(map(lambda x: int(x), flask_wtf.__version__.split('.')))

    if version < (0, 14, 0):
        return

    @app.teardown_request
    def cleanup_csrf_cache(error):
        if 'csrf_token' in flask.g:
            del flask.g.csrf_token



from __future__ import absolute_import

import flask

from keg.extensions import lazy_gettext as _


keg = flask.Blueprint('keg', __name__)


@keg.route('/exception-test')
def exception_test():
    raise Exception(_('Deliberate exception for testing purposes'))


@keg.route('/ping')
def ping():
    return '{} ok'.format(flask.current_app.name)



import keyring
import keg


try:
    import unittest.mock as mock
except ImportError:
    import mock


class MockKeyring(keyring.backend.KeyringBackend):
    priority = 1

    def __init__(self):
        self.passwords = dict()

    def set_password(self, servicename, username, password):
        key = '{}:{}'.format(servicename, username)
        self.passwords[key] = password

    def get_password(self, servicename, username):
        key = '{}:{}'.format(servicename, username)
        return self.passwords.get(key, username)

    def delete_password(self, servicename, username, password):
        pass


class TestKeyringManager:

    class MockApp:
        logger = mock.MagicMock()
        name = 'fake'

    def setup_method(self, _):
        self.manager = self.create_manager()

    def create_manager(self, app=MockApp, backend=MockKeyring):
        keyring.set_keyring(backend())
        return keg.keyring.Manager(app())

    def test_pattern_to_password(self):
        func = self.manager.pattern_to_password

        assert func('value') == 'value'
        assert func('${value}$') == 'value'
        assert func(b'${value}$') == b'value'
        assert func(b'${value}$:${other}$') == b'value:other'

    def test_substitute_finds_all_values_for_sub(self):
        data = {
            'no_ring': 'foo',
            'int_val': 1,
            'bool_val': True,
            'string_ring': '${string}$',
            'byte_ring': b'${bytes}$',
            'multi_string': '${multi_s_one}$:${multi_s_two}$',
            'multi_byte': b'${multi_b_one}$:${multi_b_two}$',
        }
        returned = self.manager.substitute(data)

        assert self.manager.sub_keys_seen == {'string', 'bytes', 'multi_s_one',
                                              'multi_s_two', 'multi_b_one',
                                              'multi_b_two'}
        assert returned['multi_byte'] == b'multi_b_one:multi_b_two'
        assert data == returned

    def test_ask_and_create(self):
        self.manager.secure_entry = lambda msg: 'password'

        self.manager.ask_and_create('key')
        assert keyring.get_password('fake', 'key') == 'password'

    def test_no_value_in_key_ring(self):
        class NoValueMockKeyring(MockKeyring):
            def get_password(self, servicename, username):
                return None

        manager = self.create_manager(backend=NoValueMockKeyring)
        manager.secure_entry = lambda msg: 'thing'

        data = manager.substitute(
            {'unknown': '${unknown}$'}
        )

        assert data['unknown'] == 'thing'



from __future__ import absolute_import

import click
import click.testing
import flask
import six
from flask_webtest import TestApp

from keg import current_app
from keg.utils import app_environ_get


def _config_profile(appcls):
    config_profile = 'TestProfile'
    if appcls.import_name is not None:
        config_profile = app_environ_get(appcls.import_name, 'CONFIG_PROFILE', config_profile)
    return config_profile


class ContextManager(object):
    """
        Facilitates having a single instance of an application ready for testing.
    """
    apps = {}

    def __init__(self, appcls):
        self.appcls = appcls
        self.app = None
        self.ctx = None

    def ensure_current(self, config):

        if not self.app:
            self.app = self.appcls().init(use_test_profile=True, config=config)
            self.ctx = self.app.app_context()
            self.ctx.push()

        if current_app._get_current_object is not self.app:
            self.ctx.push()

        return self.app

    def cleanup(self):
        self.ctx.pop()

    def is_ready(self):
        return self.app is not None

    @classmethod
    def get_for(cls, appcls):
        """
            Return the ContextManager instance for the given app class.  Only one ContextManager
            instance will be created in a Python process for any given app.
        """
        if appcls not in cls.apps:
            cls.apps[appcls] = cls(appcls)
        return cls.apps[appcls]


def invoke_command(app_cls, *args, **kwargs):
    exit_code = kwargs.pop('exit_code', 0)
    runner = kwargs.pop('runner', None) or click.testing.CliRunner()
    use_test_profile = kwargs.pop('use_test_profile', True)
    env = kwargs.pop('env', {})

    if use_test_profile:
        app_key = app_cls.environ_key('USE_TEST_PROFILE')
        env[app_key] = 'true'
    result = runner.invoke(app_cls.cli, args, env=env, catch_exceptions=False, **kwargs)

    # if an exception was raised, make sure you output the output to make debugging easier
    # -1 as an exit code indicates a non SystemExit exception.
    # 9/16/15: Not sure this is necessary anymore since we force catch_exceptions=False above (RLS).
    if result.exit_code == -1:
        click.echo(result.output)
        six.reraise(result.exc_info[1], None, result.exc_info[2])

    error_message = 'Command exit code {}, expected {}.  Result output follows:\n{}'
    assert result.exit_code == exit_code, error_message.format(result.exit_code, exit_code,
                                                               result.output)
    return result


class CLIBase(object):
    # child classes will need to set these values in order to use the class
    app_cls = None
    cmd_name = None

    @classmethod
    def setup_class(cls):
        cls.runner = click.testing.CliRunner()

    def invoke(self, *args, **kwargs):
        cmd_name = kwargs.pop('cmd_name', self.cmd_name)
        if cmd_name is None:
            cmd_name_args = []
        else:
            cmd_name_args = cmd_name.split(' ')
        invoke_args = cmd_name_args + list(args)
        kwargs['runner'] = self.runner

        # If the app_cls isn't set, use the class of the current app
        app_cls = self.app_cls or flask.current_app._get_current_object().__class__

        return invoke_command(app_cls, *invoke_args, **kwargs)


class WebBase(object):
    db = None
    appcls = None

    @classmethod
    def setup_class(cls):
        cls.app = cls.appcls.testing_prep()
        cls.testapp = TestApp(flask.current_app)



from __future__ import absolute_import

import getpass
import re
import sys
import itertools

import six

try:
    import keyring
except ImportError:
    keyring = None

from keg.extensions import lazy_gettext as _


class KeyringError(Exception):
    pass


class MissingValueError(Exception):

    def __init__(self, key):
        self.key = key

        super(Exception, self).__init__(_("No value for {key} found in key "
                                          "ring", key=key))


class Manager(object):

    # regex that matches "${*}$" where * = any printable ASCII character
    # that isn't "}"
    # see: http://www.catonmat.net/blog/my-favorite-regex/
    sub_pattern = r'(\$\{([ -|~]+?)\}\$)'

    backend_min_priority = 1

    def __init__(self, app, service_name=None, secure_entry=getpass.getpass):
        self.sub_re = re.compile(self.sub_pattern)
        self.log = app.logger
        self.app = app
        self.service_name = service_name or self.app.name
        self.sub_keys_seen = set()
        self.secure_entry = secure_entry

        if not self.verify_backend():
            self.log.warning(_(
                'Insecure keyring backend detected, keyring substitution unavailable. '
                'Run this app\'s keyring command for more info.'
            ))

    def verify_backend(self):
        backend = keyring.get_keyring()
        if backend.priority < self.backend_min_priority:
            return False
        return True

    def pattern_to_password(self, value, service_name=None):
        """
        Substitutes a configuration value which matches self.sub_pattern with
        the value from the keyring.

        :param value: the substitutable value '${value}$'
        :param service_name: a configured keyring service name
        """
        if not isinstance(value, (six.binary_type, six.text_type)):
            return value

        # preserve the original type so that we can convert at the end
        is_bytes = isinstance(value, six.binary_type)
        service = service_name or self.service_name

        # Convert to a regular string if it is bytes else use as is
        value = value.decode() if is_bytes else value

        # Find all the possible keys we need to look up. Some values need
        # multiple replacements, for example '${user}$:${pass}$'. Therefore
        # this loop will run once for each match i.e. user and pass
        for match in self.sub_re.finditer(value):
            raw = match.group(1)  # '${user}$'
            clean = match.group(2)  # user

            # TODO (NZ): I am not really wild about this being here, but I don't
            # have a good way of keeping the data without it.
            self.sub_keys_seen.add(clean)

            if self.log:
                msg = _('Keyring Substitute: replacing {raw} for key {clean}',
                        raw=raw, clean=clean)
                self.log.debug(msg)

            stored_value = keyring.get_password(service, clean)
            if not stored_value:
                raise MissingValueError(clean)

            value = value.replace(raw, stored_value, 1)

        return value.encode() if is_bytes else value

    def substitute(self, data):
        if not self.verify_backend():
            return

        def key_and_value(key, value):
            try:
                password = self.pattern_to_password(value)
            except MissingValueError as err:
                # Backwards comp... Ask for the value if we can't find it.
                #
                # TODO: Remove the need to ask for a password in this method
                # instead the substitute should work on known values and setting
                # values should be a different function
                password = self.ask_and_create(err.key)

            return key, password

        # Rifle through the dictionary, substitute where necessary
        subbed = {key: value for key, value in
                  itertools.starmap(key_and_value, six.iteritems(data))}

        # Mutate the data passed in for backwards comp.
        # TODO: Remove this mutation and just return the data allowing the
        # caller to decide what to do with it.
        data.update(subbed)
        return subbed

    def delete(self, key):
        keyring.delete_password(self.app.name, key)

    def create(self, key, value, service_name=None):
        keyring.set_password(service_name or self.service_name, key, value)

    def ask_and_create(self, key, service_name=None):
        msg = _('Missing secret for "{key}": ', key=key)

        # Windows getpass implementation breaks in Python2 when passing it Unicode
        if sys.version_info < (3,) and sys.platform.startswith('win'):
            msg = str(msg)

        value = self.secure_entry(msg)
        self.create(key, value, service_name)

        return value



from __future__ import absolute_import

from collections import defaultdict
from contextlib import contextmanager
import platform

import click
import flask
import flask.cli
from six.moves import urllib

from keg import current_app
from keg.extensions import gettext as _
from keg.keyring import keyring as keg_keyring


class KegAppGroup(flask.cli.AppGroup):
    def __init__(self, create_app, add_default_commands=True, *args, **kwargs):
        self.create_app = create_app

        flask.cli.AppGroup.__init__(self, *args, **kwargs)
        if add_default_commands:
            self.add_command(dev_command)

        self._loaded_plugin_commands = False

    def _load_plugin_commands(self):
        if self._loaded_plugin_commands:
            return
        try:
            import pkg_resources
        except ImportError:
            self._loaded_plugin_commands = True
            return

        for ep in pkg_resources.iter_entry_points('flask.commands'):
            self.add_command(ep.load(), ep.name)
        for ep in pkg_resources.iter_entry_points('keg.commands'):
            self.add_command(ep.load(), ep.name)
        self._loaded_plugin_commands = True

    def list_commands(self, ctx):
        self._load_plugin_commands()

        info = ctx.ensure_object(flask.cli.ScriptInfo)
        info.load_app()
        rv = set(click.Group.list_commands(self, ctx))
        return sorted(rv)

    def get_command(self, ctx, name):
        info = ctx.ensure_object(flask.cli.ScriptInfo)
        info.load_app()
        return click.Group.get_command(self, ctx, name)

    def main(self, *args, **kwargs):
        obj = kwargs.get('obj')
        if obj is None:
            obj = flask.cli.ScriptInfo(create_app=self.create_app)
        kwargs['obj'] = obj
        # TODO: figure out if we want to use this next line.
        #kwargs.setdefault('auto_envvar_prefix', 'FLASK')
        return flask.cli.AppGroup.main(self, *args, **kwargs)


@click.group('develop', help=_('Developer info and utils.'))
def dev_command():
    pass


dev_command.add_command(flask.cli.run_command)
dev_command.add_command(flask.cli.shell_command)


@dev_command.command('routes', short_help=_('List the routes defined for this app.'))
@flask.cli.with_appcontext
def routes_command():
    output = []
    endpoint_len = 0
    methods_len = 0

    # calculate how wide the output columns need to be
    for rule in flask.current_app.url_map.iter_rules():
        methods = ','.join(rule.methods)
        endpoint_len = max(endpoint_len, len(rule.endpoint))
        methods_len = max(methods_len, len(methods))

    # generate the output
    for rule in flask.current_app.url_map.iter_rules():
        methods = ','.join(rule.methods)
        line = urllib.parse.unquote("{}   {}   {}".format(
            rule.endpoint.ljust(endpoint_len), methods.ljust(methods_len), rule))
        output.append(line)

    for line in sorted(output):
        click.echo(line)


@dev_command.command('templates', short_help=_('Show paths searched for a template.'))
@flask.cli.with_appcontext
def templates_command():
    jinja_loader = flask.current_app.jinja_env.loader
    paths = defaultdict(list)
    for template in jinja_loader.list_templates():
        for loader, template in jinja_loader._iter_loaders(template):
            for dpath in loader.searchpath:
                paths[dpath].append(template)
    for dpath in sorted(paths.keys()):
        template_paths = sorted(paths[dpath])
        click.echo(dpath)
        for tpath in template_paths:
            click.echo('    {}'.format(tpath))


@dev_command.command('config', short_help=_('List info related to config files, profiles, and'
                     ' values.'))
@flask.cli.with_appcontext
def config_command():
    app = flask.current_app
    config = app.config
    keys = list(config.keys())
    keys.sort()

    click.echo(_('Default config objects:'))
    for val in config.default_config_locations_parsed():
        click.echo('    {}'.format(val))

    click.echo(_('Config file locations:'))
    for val in config.config_file_paths():
        click.echo('    {}'.format(val))

    click.echo(_('Config objects used:'))
    for val in config.configs_found:
        click.echo('    {}'.format(val))

    click.echo(_('Resulting app config (including Flask defaults):'))
    for key in keys:
        click.echo('    {} = {}'.format(key, config[key]))


class DatabaseGroup(click.MultiCommand):

    def list_commands(self, ctx):
        return ['clear', 'init']

    def get_command(self, ctx, name):
        if name == 'init':
            return database_init
        if name == 'clear':
            return database_clear


@dev_command.command('db', cls=DatabaseGroup, invoke_without_command=True,
                     help=_('Lists database related sub-commands.'))
@flask.cli.with_appcontext
@click.pass_context
def database_group(ctx):
    # only take action if no subcommand is involved.
    if ctx.invoked_subcommand is None:
        if not current_app.db_enabled:
            click.echo(_('Database not enabled for this app.  No subcommands available.'))
        else:
            # Database enabled, but no subcommand was given, therefore we want to just show
            # the help message, which would be the default behavior if we had not used the
            # invoke_Without_command option.
            click.echo(ctx.get_help())
            ctx.exit()


@click.command('init', short_help=_('Create all db objects, send related events.'))
@click.option('--clear-first', default=False, is_flag=True,
              help=_('Clear DB of all data and drop all objects before init.'))
@flask.cli.with_appcontext
def database_init(clear_first):
    if clear_first:
        current_app.db_manager.db_init_with_clear()
        click.echo(_('Database cleared and initialized'))
    else:
        current_app.db_manager.db_init()
        click.echo(_('Database initialized'))


@click.command('clear', short_help=_('Clear DB of all data and drop all objects.'))
@flask.cli.with_appcontext
def database_clear():
    current_app.db_manager.db_clear()
    click.echo(_('Database cleared'))


class KeyringGroup(click.MultiCommand):

    def list_commands(self, ctx):
        if keg_keyring:
            return ['delete', 'list-keys', 'setup', 'status']
        else:
            return ['status']

    def get_command(self, ctx, name):
        if name == 'status':
            return keyring_status
        if name == 'list-keys':
            return keyring_list_keys
        if name == 'delete':
            return keyring_delete


def keyring_notify_no_module():
    click.echo(_('Keyring module not installed. Keyring functionality disabled.\n\nYou can'
                 ' enable keyring functionality by installing the package:'
                 ' `pip install keyring`.'))


@dev_command.command('keyring', cls=KeyringGroup, invoke_without_command=True,
                     help=_('Lists keyring related sub-commands.'))
@click.pass_context
def keyring_group(ctx):
    # only take action if no subcommand is involved.
    if ctx.invoked_subcommand is None:
        if keg_keyring is None:
            keyring_notify_no_module()
        else:
            # keyring is available, but no subcommand was given, therefore we want to just show
            # the help message, which would be the default behavior if we had not used the
            # invoke_Without_command option.
            click.echo(ctx.get_help())
            ctx.exit()


@click.command('status', short_help=_('Show keyring related status info.'))
@click.option('--unavailable', default=False, is_flag=True,
              help=_('Show unavailable backends with reasons.'))
@flask.cli.with_appcontext
def keyring_status(unavailable):
    if keg_keyring is None:
        keyring_notify_no_module()
        return
    import keyring
    import keyring.backend as kb
    viable = kb.get_all_keyring()

    # call get_all_keyring() before this so we are sure all keyrings are loaded
    # on KeyringBackend
    if unavailable:
        click.echo(_('Unavailable backends'))
        for cls in kb.KeyringBackend._classes:
            try:
                cls.priority
            except Exception as e:
                click.echo(_('    {_class.__module__}:{_class.__name__} - {exception}',
                             _class=cls, exception=e))

    click.echo(_('\nAvailable backends (backends with priority < 1 are not'
                 ' recommended and may be insecure)'))
    for backend in viable:
        click.echo(_('    {_class.__module__}:{_class.__name__} (priority: {priority})',
                     _class=backend.__class__, priority=backend.priority))

    click.echo(_('\nDefault backend'))
    backend = keyring.get_keyring()
    click.echo(_('    {_class.__module__}:{_class.__name__}', _class=backend.__class__))
    if hasattr(backend, 'file_path'):
        click.echo(_('    file path: {file_path}', file_path=backend.file_path))

    if not flask.current_app.keyring_enabled:
        click.echo(_('\nKeyring functionality for this app has been DISABLED through the config'
                     ' setting KEG_KEYRING_ENABLE.'))
    elif not flask.current_app.keyring_manager.verify_backend():
        click.echo(_('\nWARNING: the current backend is insecure,'
                     ' keyring substitution unavailable.'))
        if platform.system() == 'Linux':
            click.echo(_('\nTRY THIS: use the SecretStorage Setup utility to get a more secure'
                         ' keyring backend.'))
            click.echo('https://pypi.python.org/pypi/SecretStorage-Setup\n')


@click.command('list-keys', short_help=_('Show all keys used in config value substitution.'))
@flask.cli.with_appcontext
def keyring_list_keys():
    km = flask.current_app.keyring_manager
    for key in sorted(km.sub_keys_seen):
        click.echo(key)


@click.command('delete', short_help=_('Delete an entry from the keyring.'))
@click.argument('key')
@flask.cli.with_appcontext
def keyring_delete(key):
    flask.current_app.keyring_manager.delete(key)


class CLILoader(object):
    """
        This loader takes care of the complexity of click object setup and instantiation in the
        correct order so that application level CLI options are available before the Keg app is
        instantiated (so the options can be used to configure the app).

        The order of events is:

        - instantiate KegAppGroup
        - KegAppGroup.main() is called
            1. the ScriptInfo object is instantiated
            2. normal click behavior starts, including argument parsing
                - arguments are always parsed due to invoke_without_command=True
            3. during argument parsing, option callbacks are excuted which will result in at least
               the profile name being saved in ScriptInfo().data
            4. Normal click .main() behavior will continue which could include processing commands
               decorated with flask.cli.with_appcontext() or calls to KegAppGroup.list_commands()
               or KegAppGroup.get_command().
                - ScriptInfo.init_app() will be called during any of these operations
                - ScriptInfo.init_app() will call self.create_app() below with the ScriptInfo
                  instance
    """
    def __init__(self, appcls):
        self.appcls = appcls
        # Don't store instance-level vars here.  This object is only instantiated once per Keg
        # sub-class but can be used across multiple app instance creations.  So, anything app
        # instance specific should go on the ScriptInfo instance (see self.options_callback()).

    def create_group(self):
        """ Create the top most click Group instance which is the entry point for any Keg app
            being called in a CLI context.

            The return value of this context gets set on Keg.cli
        """

        return KegAppGroup(
            self.create_app,
            params=self.create_script_options(),
            callback=self.main_callback,
            invoke_without_command=True,
        )

    def create_app(self, script_info):
        """ Instantiate our app, sending CLI option values through as needed. """
        init_kwargs = self.option_processor(script_info.data)
        return self.appcls().init(**init_kwargs)

    def option_processor(self, cli_options):
        """
            Turn cli_options, which is the result of all parsed options in create_script_options()
            into a dict which will be given as kwargs to the app's .init() method.
        """
        retval = dict(config_profile=cli_options.get('profile', None), config={})
        if cli_options.get('quiet', False):
            retval['config']['KEG_LOG_STREAM_ENABLED'] = False
        return retval

    def create_script_options(self):
        """ Create app level options, ideally that are used to configure the app itself.  """
        return [
            click.Option(['--profile'], is_eager=True, default=None, callback=self.options_callback,
                         help=_('Name of the configuration profile to use.'),),
            click.Option(['--quiet'], is_eager=True, is_flag=True, default=False,
                         callback=self.options_callback,
                         help=_('Set default logging level to logging.WARNING.')),
            click.Option(['--help-all'], is_eager=True, is_flag=True, expose_value=False,
                         callback=self.help_all_callback,
                         help=_('Show all commands with subcommands.')),
        ]

    def options_callback(self, ctx, param, value):
        """ This method is called after argument parsing, after ScriptInfo instantiation but before
            create_app() is called.  It's the only way to get the options into ScriptInfo.data
            before the Keg app instance is instantiated.
        """
        si = ctx.ensure_object(flask.cli.ScriptInfo)
        si.data[param.name] = value

    def help_all_callback(self, ctx, param, value):
        if not value or ctx.resilient_parsing:
            return
        formatter = ctx.make_formatter()

        ctx.command.format_usage(ctx, formatter)
        ctx.command.format_help_text(ctx, formatter)
        self.format_options(ctx.command, ctx, formatter)
        self.format_commands(ctx.command, ctx, formatter)
        ctx.command.format_epilog(ctx, formatter)

        click.echo(formatter.getvalue().rstrip('\n'))
        ctx.exit()

    def format_options(self, command, ctx, formatter):
        opts = []
        for param in command.get_params(ctx):
            if param.name == 'help':
                opts.append(('--help', _('Show help message.')))
            elif param.name == 'help_all':
                opts.append(('--help-all', _('Show this message and exit.')))
            else:
                rv = param.get_help_record(ctx)
                if rv is not None:
                    opts.append(rv)

        if opts:
            with formatter.section('Options'):
                formatter.write_dl(opts)

    def format_commands(self, command, ctx, formatter):
        if hasattr(command, 'list_commands'):
            subcommands = command.list_commands(ctx)

            if len(subcommands):
                # allow for 3 times the default spacing
                limit = formatter.width - 6 - max(len(subcommand) for subcommand in subcommands)

                rows = []
                commands = []
                for subcommand in subcommands:
                    cmd = command.get_command(ctx, subcommand)
                    if cmd is None:
                        continue
                    if cmd.hidden:
                        continue
                    commands.append(cmd)
                    help = cmd.get_short_help_str(limit)
                    rows.append((subcommand, help))

                with self.compact_section(formatter, 'Commands'):
                    self.write_dl_with_subcommands(ctx, formatter, rows, commands)

    def write_dl_with_subcommands(self, ctx, formatter, rows, commands, col_max=30, col_spacing=2):
        rows = list(rows)
        widths = click.formatting.measure_table(rows)
        if len(widths) != 2:
            raise TypeError('Expected two columns for definition list')

        first_col = min(widths[0], col_max) + col_spacing

        for (first, second), command in zip(
            click.formatting.iter_rows(rows, len(widths)), commands
        ):
            formatter.write('%*s%s' % (formatter.current_indent, '', first))
            if not second:
                formatter.write('\n')
                continue
            if click.formatting.term_len(first) <= first_col - col_spacing:
                formatter.write(' ' * (first_col - click.formatting.term_len(first)))
            else:
                formatter.write('\n')
                formatter.write(' ' * (first_col + formatter.current_indent))

            text_width = max(formatter.width - first_col - 2, 10)
            lines = iter(click.formatting.wrap_text(second, text_width).splitlines())
            if lines:
                formatter.write(next(lines) + '\n')
                for i, line in enumerate(lines):
                    formatter.write('%*s%s\n' % (
                        first_col + formatter.current_indent, '', line))
            else:
                formatter.write('\n')

            # Add subcommands under current command (recursive)
            with formatter.indentation():
                self.format_commands(command, ctx, formatter)

    @contextmanager
    def compact_section(self, formatter, name):
        formatter.write_heading(name)
        formatter.indent()
        try:
            yield
        finally:
            formatter.dedent()

    def main_callback(self, **kwargs):
        """
            Default Click behavior is to call the help method if no arguments are given to the
            top-most command.  That's good UX, but the way Click impliments it, the help is
            shown before argument parsing takes place.  That's bad, because it means our
            options_callback() is never called and the config profile isn't available when the
            help command calls KegAppGroup.list_commands().

            So, we force Click to always parse the args using `invoke_without_command=True` above,
            but when we do that, Click turns off the automatic display of help.  So, we just
            impliment help-like behavior in this method, which gives us the same net result
            as default Click behavior.
        """
        ctx = click.get_current_context()

        if ctx.invoked_subcommand is not None:
            # A subcommand is present, so arguments were passed.
            return

        click.echo(ctx.get_help(), color=ctx.color)
        ctx.exit()



from __future__ import absolute_import

import tempfile

from keg.utils import pymodule_fpaths_to_objects


class TestUtils(object):

    def test_pymodule_fpaths_to_objects(self):
        with tempfile.NamedTemporaryFile(delete=False) as fh:
            fpath1 = fh.name
            fh.write(b'foo1="bar"')

        with tempfile.NamedTemporaryFile(delete=False) as fh:
            fpath2 = fh.name
            fh.write(b'foo2="bar"')

        with tempfile.NamedTemporaryFile() as fh:
            fpath3 = fh.name

        result = pymodule_fpaths_to_objects([fpath1, fpath2, fpath3])

        fpath, fpath_objs = result.pop(0)
        assert fpath == fpath1
        assert fpath_objs['foo1'] == 'bar'

        fpath, fpath_objs = result.pop(0)
        assert fpath == fpath2
        assert fpath_objs['foo2'] == 'bar'

        # result should now be empty since fpath3 should get filtered out b/c it's not present
        assert len(result) == 0



from __future__ import absolute_import

from flask import url_for, session, request, jsonify, Blueprint, flash, current_app
from flask_oauthlib.client import OAuth

from keg.extensions import lazy_gettext as _
from keg.web import redirect

oauthlib = OAuth()


class OAuthError(Exception):
    pass


class Provider(object):
    def authorize(self, **kwargs):
        return self.remote_app.authorize(**kwargs)

    def get(self, *args):
        return self.remote_app.get(*args)


class Google(Provider):
    def __init__(self):
        self.remote_app = oauthlib.remote_app(
            'google',
            request_token_params={
                'scope': 'https://www.googleapis.com/auth/userinfo.email'
            },
            base_url='https://www.googleapis.com/oauth2/v1/',
            request_token_url=None,
            access_token_method='POST',
            access_token_url='https://accounts.google.com/o/oauth2/token',
            authorize_url='https://accounts.google.com/o/oauth2/auth',
            app_key='KEG_OAUTH_GOOGLE',
        )
        self.remote_app.tokengetter(manager.token)

    def authorized_response(self):
        resp = self.remote_app.authorized_response()

        if isinstance(resp, Exception):
            flash(str(resp), 'error')
        elif resp is None:
            error_message = _('authentication failure: reason={reason} error={error}',
                              reason=request.args.get('error_reason', _('unknown')),
                              error=request.args.get('error', _('unknown')))
            flash(error_message, 'error')
        else:
            return resp['access_token']


class Manager(object):
    provider_classes = {
        'google': Google,
    }

    def __init__(self):
        self.providers = {}

    def register_providers(self, providers):
        for provider in providers:
            if provider not in self.provider_classes:
                # if it's not a string of a known provider, then assume its a Provider instance
                self.providers[provider.name] = provider()
            else:
                self.providers[provider] = self.provider_classes[provider]()

    def session_key(self, key):
        return '{}.oauth.{}'.format(current_app.name, key)

    def provider(self, provider_key=None):
        provider_key = provider_key or session[self.session_key('provider')]
        try:
            return self.providers[provider_key]
        except KeyError:
            raise OAuthError(_('Provider "{provider}" has not been registered.',
                               provider=provider_key))

    def authorize(self, provider, callback):
        session[self.session_key('provider')] = provider
        return self.provider(provider).authorize(callback=callback)

    def authorized_response(self):
        provider = session[self.session_key('provider')]
        token = self.provider(provider).authorized_response()
        if token:
            session[self.session_key('token')] = token

    def authenticated(self):
        return bool(session.get(self.session_key('token'), None))

    def get(self, *args):
        return self.provider().get(*args)

    def userinfo(self):
        return self.provider().userinfo()

    def clear(self):
        token_key = self.session_key('token')
        if token_key in session:
            del session[token_key]

    def token(self):
        return session[self.session_key('token')]


manager = Manager()

bp = Blueprint('oauth', __name__, url_prefix='/oauth')


@bp.route('/userinfo')
def userinfo():
    if manager.authenticated():
        return jsonify(manager.userinfo())
    return redirect('LOGIN')


@bp.route('/login/<provider>')
def login(provider):
    callback_url = url_for('.callback', _external=True)
    return manager.authorize(provider, callback_url)


@bp.route('/logout')
def logout():
    manager.clear()
    return redirect('AFTER_LOGOUT')


@bp.route('/callback')
def callback():
    manager.authorized_response()
    return redirect('AFTER_LOGIN')



from __future__ import absolute_import

import flask_sqlalchemy as fsa

from keg.signals import testing_run_start, db_init_pre, db_init_post, db_clear_pre, db_clear_post
from keg.utils import visit_modules


class KegSQLAlchemy(fsa.SQLAlchemy):

    def apply_driver_hacks(self, app, info, options):
        """Override some driver specific settings"""
        super(KegSQLAlchemy, self).apply_driver_hacks(app, info, options)

        engine_opts = app.config.get('KEG_DB_ENGINE_OPTIONS', {})

        # Turn on SA pessimistic disconnect handling by default:
        # http://docs.sqlalchemy.org/en/latest/core/pooling.html#disconnect-handling-pessimistic
        engine_opts.setdefault('pool_pre_ping', True)

        # Update the options with the engine options we received from the application
        options.update(engine_opts)

    def get_engines(self, app):
        # the default engine doesn't have a bind
        retval = [(None, self.get_engine(app))]

        bind_names = app.config['SQLALCHEMY_BINDS']
        # The default value of SQLALCHEMY_BINDS is None and the key is present b/c of
        # Flask-SQLAlchemy defaults.  So, only process the binds if they are not None.
        if bind_names is not None:
            for bind_name in bind_names:
                retval.append((bind_name, self.get_engine(app, bind=bind_name)))

        return retval


db = KegSQLAlchemy()

# put this import after the above db assignment to avoid circular reference issues
from .dialect_ops import DialectOperations  # noqa


class DatabaseManager(object):
    """
        A per-app-instance utility class that managers all common operations for a Keg app.

        Binds & Dialects
        ----------------

        Flask SQLAlchemy handles multiple DB connections per application through the use of "binds."
        When an application wants to communicate events or initiate activites, this manager will
        will handle distributing those events and activities to all database connections bound
        to the application.

        Furthermore, this manager delegates to DialectOperations instances to run the events and
        activities in ways that are specific to the type of RDBMS being used (when needed).
    """

    def __init__(self, app):
        self.app = app
        self.dialect_opts = app.config['KEG_DB_DIALECT_OPTIONS']

        self.init_app()
        self.init_events()

    def init_app(self):
        db.init_app(self.app)
        visit_modules(self.app.db_visit_modules, self.app.import_name)

    def init_events(self):
        testing_run_start.connect(self.on_testing_start, sender=self.app)

    def bind_dialect(self, bind_name):
        engine = db.get_engine(self.app, bind=bind_name)
        return DialectOperations.create_for(engine, bind_name, self.dialect_opts)

    def all_bind_dialects(self):
        """
            For each database connection (bind) in this application, yield a DialectOperations
            instance corresponding to the type of RDBMS the bind is connecting to.
        """
        for bind_name, engine in db.get_engines(self.app):
            yield DialectOperations.create_for(engine, bind_name, self.dialect_opts)

    def on_testing_start(self, app):
        self.db_init_with_clear()

    def drop_all(self):
        db.session.remove()
        for dialect in self.all_bind_dialects():
            dialect.drop_all()

    def prep_empty(self):
        for dialect in self.all_bind_dialects():
            dialect.prep_empty()

    # The methods that follow will trigger application events.
    def db_init_with_clear(self):
        self.db_clear()
        self.db_init()

    def db_init(self):
        db_init_pre.send(self.app)
        db.create_all()
        db_init_post.send(self.app)

    def db_clear(self):
        db_clear_pre.send(self.app)
        self.drop_all()
        # todo: prep_empty() should probably be an event
        self.prep_empty()
        db_clear_post.send(self.app)




class DefaultProfile(object):
    KEG_KEYRING_ENABLE = False

    KEG_REGISTERED_COMPONENTS = {
        'keg_apps.db.blog',
    }


class TestProfile(object):
    KEG_DB_DIALECT_OPTIONS = {
        'postgresql.schemas': ('public', 'fooschema'),
        'mssql.schemas': ('fooschema', ),
    }

    # These binds are used to test keg dialect functionality in CircleCI. All other database tests
    # are done on the primary connection, which defaults to SQLite.
    #
    # Appveyor: see user-config-tpl-appveyor.py
    # Developers: see instructions in readme for setup of local config file
    SQLALCHEMY_BINDS = {
        'postgres': 'postgresql://postgres:password@localhost/postgres',
        'sqlite2': 'sqlite:///'
    }



from __future__ import absolute_import

import six

from blazeutils.strings import simplify


def object_from_source(fpath, objname):
    if six.PY2:
        import imp
        module_name = 'keg_{}'.format(simplify(fpath))
        tmpmodule = imp.load_source(module_name, fpath)
        if not hasattr(tmpmodule, objname):
            return None
        return getattr(tmpmodule, objname)



from __future__ import absolute_import

from keg.testing import WebBase

from keg_apps.web.app import WebApp


class TestBaseViewFeatures(WebBase):
    appcls = WebApp

    def test_auto_assign(self):
        resp = self.testapp.get('/auto-assign')
        lines = resp.body.splitlines()
        assert lines[0] == b'bar = bar'
        assert lines[1] == b'baz = baz'
        assert lines[2] == b'foo = none'

    def test_no_auto_assign_with_response(self):
        resp = self.testapp.get('/auto-assign-with-response')
        lines = resp.body.splitlines()
        assert lines[0] == b'bar = '


class TestBlueprintUsage(WebBase):
    appcls = WebApp

    def test_blueprint_url_prefix(self):
        self.testapp.get('/tanagra/blueprint-test')

    def test_blueprint_template_folder(self):
        resp = self.testapp.get('/tanagra/blueprint-test')
        assert 'blueprint template found ok' in resp


class TestComponentBlueprint(WebBase):
    appcls = WebApp

    def test_component_blueprint_loads(self):
        resp = self.testapp.get('/blog')
        assert 'I am a blog' in resp
